@@ -4,6 +4,7 @@ OMFIT
 
 """
 
+from collections import OrderedDict
 import copy
 import glob
 import io
@@ -11,14 +12,25 @@ import numpy
 import os
 import re
 
-from boutdata.collect import collect, create_cache
+from multiprocessing import Process, Pipe, RawArray
+
+from boutdata.collect import (
+    collect,
+    create_cache,
+    findVar,
+    _check_fieldperp_attributes,
+    _collect_from_one_proc,
+    _get_grid_info,
+)
+from boututils.boutarray import BoutArray
 from boututils.boutwarnings import alwayswarn
 from boututils.datafile import DataFile
+from boututils.run_wrapper import determineNumberOfCPUs
 
 # These are imported to be used by 'eval' in
 # BoutOptions.evaluate_scalar() and BoutOptionsFile.evaluate().
 # Change the names to match those used by C++/BOUT++
-from numpy import (
+from numpy import (  # noqa: F401
     pi,
     sin,
     cos,
@@ -211,9 +223,7 @@ class BoutOptions(object):
         return self._pop_impl(key, default)[0]
 
     def _pop_impl(self, key, default=__marker):
-        """Private implementation of pop; also pops metadata
-
-        """
+        """Private implementation of pop; also pops metadata"""
         key_parts = key.split(":", maxsplit=1)
 
         if len(key_parts) > 1:
@@ -239,8 +249,7 @@ class BoutOptions(object):
         return (value, name, parent, comment, inline_comment, comment_whitespace)
 
     def rename(self, old_name, new_name):
-        """Rename old_name to new_name
-        """
+        """Rename old_name to new_name"""
 
         def setattr_nested(parent, key, attr, value):
             """Set one of the comment types on some nested section. Slightly
@@ -258,18 +267,17 @@ class BoutOptions(object):
             else:
                 getattr(parent, attr)[key] = value
 
-        def ensure_sections(parent, path):
-            """Make sure all the components of path in parent are sections
-            """
-            path_parts = path.split(":", maxsplit=1)
-
-            def check_is_section(parent, path):
-                if path in parent and not isinstance(parent[path], BoutOptions):
-                    raise TypeError(
-                        "'{}:{}' already exists and is not a section!".format(
-                            parent._name, path
-                        )
+        def check_is_section(parent, path):
+            if path in parent and not isinstance(parent[path], BoutOptions):
+                raise TypeError(
+                    "'{}:{}' already exists and is not a section!".format(
+                        parent._name, path
                     )
+                )
+
+        def ensure_sections(parent, path):
+            """Make sure all the components of path in parent are sections"""
+            path_parts = path.split(":", maxsplit=1)
 
             if len(path_parts) > 1:
                 new_parent_name, child_name = path_parts
@@ -280,11 +288,54 @@ class BoutOptions(object):
                 check_is_section(parent, path)
                 parent.getSection(path)
 
+        def rename_key(thing, new_name, old_name):
+            """Rename a key in a dict while trying to preserve order, useful for
+            minimising diffs"""
+            return {new_name if k == old_name else k: v for k, v in thing.items()}
+
+        def get_immediate_parent_and_child(path):
+            """Get the immediate parent of path"""
+            parent, _, child = path.rpartition(":")
+            if parent and parent in self:
+                return self[parent], child
+            return self, path
+
         value = self[old_name]
 
         if isinstance(value, BoutOptions):
             # We're moving a section: make sure we don't clobber existing values
             ensure_sections(self, new_name)
+
+            new_parent, new_child = get_immediate_parent_and_child(new_name)
+            old_parent, old_child = get_immediate_parent_and_child(old_name)
+
+            # Did we just add a new section?
+            new_section = len(new_parent[new_child].keys()) == 0
+            # Was it just a change in case?
+            case_change = new_child.lower() == old_child.lower()
+
+            # Renaming a child section just within the same parent section, we can
+            # preserve the order
+            if (new_parent is old_parent) and (new_section or case_change):
+                # We just put a new section in, but it will have been
+                # added at the end -- remove it so we can actually put
+                # the new section in the same order as the original
+                if new_section:
+                    new_parent.pop(new_child)
+                new_parent._sections = rename_key(
+                    new_parent._sections, new_child, old_child
+                )
+                new_parent.comments = rename_key(
+                    new_parent.comments, new_child, old_child
+                )
+                new_parent.inline_comments = rename_key(
+                    new_parent.inline_comments, new_child, old_child
+                )
+                new_parent._comment_whitespace = rename_key(
+                    new_parent._comment_whitespace, new_child, old_child
+                )
+                return
+
             # Now we're definitely moving into an existing section, so
             # update values and comments
             for key in value:
@@ -306,6 +357,24 @@ class BoutOptions(object):
                 old_name
             )
         else:
+            new_parent, new_child = get_immediate_parent_and_child(new_name)
+            old_parent, old_child = get_immediate_parent_and_child(old_name)
+
+            # Renaming a child key just within the same parent section, we can preserve
+            # the order
+            if new_parent is old_parent:
+                new_parent._keys = rename_key(new_parent._keys, new_child, old_child)
+                new_parent.comments = rename_key(
+                    new_parent.comments, new_child, old_child
+                )
+                new_parent.inline_comments = rename_key(
+                    new_parent.inline_comments, new_child, old_child
+                )
+                new_parent._comment_whitespace = rename_key(
+                    new_parent._comment_whitespace, new_child, old_child
+                )
+                return
+
             _, _, _, comment, inline_comment, comment_whitespace = self._pop_impl(
                 old_name
             )
@@ -327,27 +396,19 @@ class BoutOptions(object):
         return self._name
 
     def keys(self):
-        """Returns all keys, including sections and values
-
-        """
+        """Returns all keys, including sections and values"""
         return list(self._sections) + list(self._keys)
 
     def sections(self):
-        """Return a list of sub-sections
-
-        """
+        """Return a list of sub-sections"""
         return self._sections.keys()
 
     def values(self):
-        """Return a list of values
-
-        """
+        """Return a list of values"""
         return self._keys.keys()
 
     def as_dict(self):
-        """Return a nested dictionary of all the options.
-
-        """
+        """Return a nested dictionary of all the options."""
         dicttree = {name: self[name] for name in self.values()}
         dicttree.update({name: self[name].as_dict() for name in self.sections()})
         return dicttree
@@ -375,18 +436,14 @@ class BoutOptions(object):
         return True
 
     def __iter__(self):
-        """Iterates over all keys. First values, then sections
-
-        """
+        """Iterates over all keys. First values, then sections"""
         for k in self._keys:
             yield k
         for s in self._sections:
             yield s
 
     def as_tree(self, indent=""):
-        """Return a string formatted as a pretty version of the options tree
-
-        """
+        """Return a string formatted as a pretty version of the options tree"""
         text = self._name + "\n"
 
         for k in self._keys:
@@ -429,6 +486,48 @@ class BoutOptions(object):
 
         return f.getvalue()
 
+    def get_bool(self, name, default=None):
+        """
+        Convert an option value to a bool, in (almost) the same way as BOUT++.
+
+        Warnings
+        --------
+        BOUT++ will convert any option value beginning with "y", "Y", "t", "T" or "1" to
+        True, and any beginning with "n", "N", "f", "F" or "0" to False. Because
+        BoutOptions converts option values to int and float, this method cannot be quite
+        so permissive, and will raise an exception for ints other than 0 and 1 and for
+        floats, which BOUT++ might convert to a bool.
+
+        Parameters
+        ----------
+        name : str
+            The name of the option to read
+        default : bool, optional
+            Value to return if the option is not present. If default is not provided an
+            exception will be raised if the option is not present.
+        """
+        if default is not None and not isinstance(default, bool):
+            raise ValueError(f'default "{default}" is not a bool')
+
+        try:
+            value = self[name]
+        except KeyError:
+            if default is None:
+                raise
+            else:
+                return default
+
+        if value == 1 or (
+            isinstance(value, str) and value.lower() in ("y", "yes", "t", "true")
+        ):
+            return True
+        elif value == 0 or (
+            isinstance(value, str) and value.lower() in ("n", "no", "f", "false")
+        ):
+            return False
+
+        raise ValueError(f"Could not convert {name}={value} to a bool")
+
     def evaluate_scalar(self, name):
         """
         Evaluate (recursively) scalar expressions
@@ -468,7 +567,8 @@ class BoutOptions(object):
             if re.search(
                 r"(?<!:)\b" + re.escape(nested_name.lower()) + r"\b", expression.lower()
             ):
-                # match nested_name only if not preceded by colon (which indicates more nesting)
+                # match nested_name only if not preceded by colon (which indicates more
+                # nesting)
                 expression = re.sub(
                     r"(?<!:)\b" + re.escape(nested_name.lower()) + r"\b",
                     "(" + self._substitute_expressions(var) + ")",
@@ -555,7 +655,7 @@ class BoutOptionsFile(BoutOptions):
             for linenr, line in enumerate(f.readlines()):
                 # First remove comments, either # or ;
                 if line.lstrip().startswith(self.VALID_COMMENTS):
-                    comments.append('#' + line.strip()[1:])
+                    comments.append("#" + line.strip()[1:])
                     continue
                 if line.strip() == "":
                     comments.append(line.strip())
@@ -564,7 +664,7 @@ class BoutOptionsFile(BoutOptions):
                 comment_match = self.COMMENT_REGEX.search(line)
                 if comment_match is not None:
                     line, comment_whitespace, inline_comment = comment_match.groups()
-                    inline_comment = '#' + inline_comment.strip()[1:]
+                    inline_comment = "#" + inline_comment.strip()[1:]
                 else:
                     inline_comment = None
                     comment_whitespace = None
@@ -636,7 +736,7 @@ class BoutOptionsFile(BoutOptions):
                 + "\nEvaluating non-scalar options not available"
             )
 
-    def recalculate_xyz(self, *,  nx=None, ny=None, nz=None):
+    def recalculate_xyz(self, *, nx=None, ny=None, nz=None):
         """
         Recalculate the x, y avd z arrays used to evaluate expressions
         """
@@ -677,9 +777,7 @@ class BoutOptionsFile(BoutOptions):
                     # get nx, ny, nz from output files
                     from boutdata.collect import findFiles
 
-                    file_list = findFiles(
-                        path=os.path.dirname("."), prefix="BOUT.dmp"
-                    )
+                    file_list = findFiles(path=os.path.dirname("."), prefix="BOUT.dmp")
                     with DataFile(file_list[0]) as f:
                         self.nx = f["nx"]
                         self.ny = f["ny"]
@@ -734,7 +832,30 @@ class BoutOptionsFile(BoutOptions):
             ]
         )
 
-    def evaluate(self, name):
+        # Also create staggered versions of the coordinates
+        self.xlow = numpy.linspace(
+            -mxg / (self.nx - 2 * mxg),
+            1.0 + (mxg - 1.0) / (self.nx - 2 * mxg),
+            self.nx,
+        )[:, numpy.newaxis, numpy.newaxis]
+        self.ylow = (
+            2.0
+            * numpy.pi
+            * numpy.linspace(
+                -myg / self.ny,
+                1.0 + (myg - 1.0) / self.ny,
+                self.ny + 2 * myg,
+            )[numpy.newaxis, :, numpy.newaxis]
+        )
+        self.zlow = (
+            2.0
+            * numpy.pi
+            * numpy.linspace(0.0, 1.0 - 1.0 / self.nz, self.nz)[
+                numpy.newaxis, numpy.newaxis, :
+            ]
+        )
+
+    def evaluate(self, name, *, location="CELL_CENTRE"):
         """Evaluate (recursively) expressions
 
         Sections and subsections must be given as part of 'name',
@@ -747,6 +868,13 @@ class BoutOptionsFile(BoutOptions):
             subsections
 
         """
+        possible_locations = ["CELL_CENTRE", "CELL_XLOW", "CELL_YLOW", "CELL_ZLOW"]
+        if location not in possible_locations:
+            raise ValueError(
+                f"Unrecognised location {location}. Should be one of "
+                f"{possible_locations}."
+            )
+
         section = self
         split_name = name.split(":")
         for subsection in split_name[:-1]:
@@ -757,15 +885,19 @@ class BoutOptionsFile(BoutOptions):
         expression = expression.replace("^", "**")
 
         # substitute for x, y and z coordinates
-        for coord in ["x", "y", "z"]:
+        for coord, coord_at_location in [
+            ("x", "x" if location != "CELL_XLOW" else "xlow"),
+            ("y", "y" if location != "CELL_YLOW" else "ylow"),
+            ("z", "z" if location != "CELL_ZLOW" else "zlow"),
+        ]:
             expression = re.sub(
-                r"\b" + coord.lower() + r"\b", "self." + coord, expression
+                r"\b" + coord.lower() + r"\b", "self." + coord_at_location, expression
             )
 
         return eval(expression)
 
     def write(self, filename=None, overwrite=False):
-        """ Write to BOUT++ options file
+        """Write to BOUT++ options file
 
         This method will throw an error rather than overwriting an existing
         file unless the overwrite argument is set to true.
@@ -823,9 +955,31 @@ class BoutOutputs(object):
         Switch for creation of a cache of DataFile objects to be
         passed to collect so that DataFiles do not need to be
         re-opened to read each variable (default: True)
+    info : bool, optional
+        Print information about grid and data loading? (default: False)
+    xguards : bool, optional
+        Collect X boundary guard cells? (default: True)
+        (Set to True to be consistent with the definition of nx)
+    yguards : bool or "include_upper", optional
+        Collect Y boundary guard cells? (default: False)
+        If yguards=="include_upper" the y-boundary cells from the upper (second) target
+        are also included.
+    xind, yind, zind, tind : int, slice or list of int, optional
+        Range of X, Y, Z or time indices to collect. Either a single
+        index to collect, a list containing [start, end] (inclusive
+        end), or a slice object (usual python indexing). Default is to
+        fetch all indices
+    parallel : bool or int, default False
+        If set to True or 0, use the multiprocessing library to read data in parallel
+        with the maximum number of available processors. If set to an int, use that many
+        processes.
+    tind_auto : bool, optional
+        Read all files, to get the shortest length of time_indices. All data truncated
+        to the shortest length.  Useful if writing got interrupted (default: False)
 
-    **kwargs
-        keyword arguments that are passed through to _caching_collect()
+    Other parameters
+    ----------------
+    keyword arguments that are passed through to collect()
 
     Examples
     --------
@@ -855,7 +1009,16 @@ class BoutOutputs(object):
         suffix=None,
         caching=False,
         DataFileCaching=True,
-        **kwargs
+        info=False,
+        xguards=True,
+        yguards=False,
+        tind=None,
+        xind=None,
+        yind=None,
+        zind=None,
+        tind_auto=False,
+        parallel=False,
+        **kwargs,
     ):
         """
         Initialise BoutOutputs object
@@ -871,8 +1034,30 @@ class BoutOutputs(object):
             # normalize suffix by removing leading '.' if present
             self._suffix = suffix.lstrip(".")
         self._caching = caching
-        self._DataFileCaching = DataFileCaching
+        self._info = info
+        self._xguards = xguards
+        self._yguards = yguards
         self._kwargs = kwargs
+        self._parallel = parallel
+        if self._parallel is False:
+            self._DataFileCaching = DataFileCaching
+        else:
+            # parallel functionality caches DataFiles in worker processes
+            self._DataFileCaching = False
+
+        if tind_auto:
+            if self._kwargs.get("tind", None) is not None:
+                raise ValueError("Cannot use 'tind' argument with 'tind_auto=True'")
+            nt = len(
+                self._collect(
+                    "t_array",
+                    path=self._path,
+                    prefix=self._prefix,
+                    tind_auto=True,
+                    **self._kwargs,
+                )
+            )
+            self._kwargs["tind"] = slice(nt)
 
         # Label for this data
         self.label = path
@@ -888,65 +1073,151 @@ class BoutOutputs(object):
         if len(self._file_list) == 0:
             raise ValueError("ERROR: No data files found")
 
-        # Available variables
-        self.varNames = []
-        self.dimensions = {}
-        self.evolvingVariableNames = []
-
         with DataFile(latest_file) as f:
-            npes = f.read("NXPE") * f.read("NYPE")
-            if len(self._file_list) != npes:
-                alwayswarn("Too many data files, reading most recent ones")
-                if npes == 1:
-                    # single output file
-                    # do like this to catch, e.g. either 'BOUT.dmp.nc' or 'BOUT.dmp.0.nc'
-                    self._file_list = [latest_file]
-                else:
-                    self._file_list = [
-                        os.path.join(
-                            path, self._prefix + "." + str(i) + "." + self._suffix
-                        )
-                        for i in range(npes)
-                    ]
+            self.grid_info, self.tind, self.xind, self.yind, self.zind = _get_grid_info(
+                f,
+                xguards=self._xguards,
+                yguards=self._yguards,
+                tind=tind,
+                xind=xind,
+                yind=yind,
+                zind=zind,
+                nfiles=len(self._file_list),
+                all_vars_info=True,
+            )
 
-            # Get variable names
-            self.varNames = f.keys()
-            for name in f.keys():
-                dimensions = f.dimensions(name)
-                self.dimensions[name] = dimensions
-                if name != "t_array" and "t" in dimensions:
-                    self.evolvingVariableNames.append(name)
+        if len(self._file_list) != self.grid_info["npes"]:
+            alwayswarn("Too many data files, reading most recent ones")
+        if (
+            len(self._file_list) != self.grid_info["npes"]
+            and self.grid_info["npes"] == 1
+        ):
+            # single output file
+            # do like this to catch, e.g. either 'BOUT.dmp.nc' or 'BOUT.dmp.0.nc'
+            self._file_list = [latest_file]
+        else:
+            # Re-create self._file_list so that it is sorted
+            self._file_list = [
+                os.path.join(path, self._prefix + "." + str(i) + "." + self._suffix)
+                for i in range(self.grid_info["npes"])
+            ]
 
-        # Private variables
-        if self._caching:
-            from collections import OrderedDict
-
-            self._datacache = OrderedDict()
-            if self._caching is not True:
-                # Track the size of _datacache and limit it to a maximum of _caching
-                try:
-                    # Check that _caching is a number of some sort
-                    float(self._caching)
-                except ValueError:
-                    raise ValueError(
-                        "BoutOutputs: Invalid value for caching argument. Caching should be either a number (giving the maximum size of the cache in GB), True for unlimited size or False for no caching."
+        if self._info:
+            print(
+                "mxsub = {} mysub = {} mz = {}\n".format(
+                    self.grid_info["mxsub"],
+                    self.grid_info["mysub"],
+                    self.grid_info["nz"],
+                )
+            )
+            print(
+                "nxpe = {}, nype = {}, npes = {}\n".format(
+                    self.grid_info["nxpe"],
+                    self.grid_info["nype"],
+                    self.grid_info["npes"],
+                )
+            )
+            if self.grid_info["npes"] < len(self._file_list):
+                print(
+                    "WARNING: More files than expected ({})".format(
+                        self.grid_info["npes"]
                     )
-                self._datacachesize = 0
-                self._datacachemaxsize = self._caching * 1.0e9
+                )
+            elif self.grid_info["npes"] > len(self._file_list):
+                print(
+                    "WARNING: Some files missing. Expected {}".format(
+                        self.grid_info["npes"]
+                    )
+                )
 
+        # Initialise private variables
+        if self._caching:
+            self._init_caching()
+        if self._parallel is not False:
+            self._init_parallel()
         self._DataFileCache = None
 
-    def keys(self):
-        """Return a list of available variable names
+    def __del__(self):
+        if self._parallel is not False:
+            self._root_file.close()
+            for worker, connection in self._workers:
+                # Send None to terminate worker process cleanly
+                connection.send(None)
+                worker.join()
+                connection.close()
 
+    def _init_caching(self):
         """
-        return self.varNames
+        Initialise private members used for caching of data variables
+        """
+        self._datacache = OrderedDict()
+        if self._caching is not True:
+            # Track the size of _datacache and limit it to a maximum of _caching
+            try:
+                # Check that _caching is a number of some sort
+                float(self._caching)
+            except ValueError:
+                raise ValueError(
+                    "BoutOutputs: Invalid value for caching argument. Caching should "
+                    "be either a number (giving the maximum size of the cache in GB), "
+                    "True for unlimited size or False for no caching."
+                )
+            self._datacachesize = 0
+            self._datacachemaxsize = self._caching * 1.0e9
+
+    def _init_parallel(self):
+        """
+        Initialise private members used for parallel reading
+        """
+        if self._parallel is True or self._parallel == 0:
+            self._parallel = determineNumberOfCPUs()
+        if not isinstance(self._parallel, int) or self._parallel <= 0:
+            raise ValueError(
+                "Passed or found inconsistent value %i for number of processes",
+                self._parallel,
+            )
+
+        if self._parallel > self.grid_info["npes"]:
+            # Using current self._parallel, some workers would have no work
+            self._parallel = self.grid_info["npes"]
+
+        # Open the 0'th file so we can read scalars without the worker processes
+        self._root_file = DataFile(self._file_list[0])
+
+        # Need to initialise all workers with a shared memory buffer to write to
+        dim_sizes = tuple(self.grid_info["sizes"][d] for d in ("t", "x", "y", "z"))
+        self._shared_buffer_raw = RawArray("d", int(numpy.prod(dim_sizes)))
+        self._shared_buffer = numpy.reshape(
+            numpy.frombuffer(self._shared_buffer_raw), dim_sizes
+        )
+
+        # Work out which files to assign to which workers
+        min_files_per_proc = int(self.grid_info["npes"]) // self._parallel
+        extra_files = int(self.grid_info["npes"]) % self._parallel
+        files_per_proc = [min_files_per_proc] * (self._parallel - extra_files) + [
+            min_files_per_proc + 1
+        ] * extra_files
+        # self._workers is a list of pairs of (worker, connection)
+        self._workers = []
+        filenum = 0
+        for i in range(self._parallel):
+            parent_connection, child_connection = Pipe()
+            proc_list = tuple(p for p in range(filenum, filenum + files_per_proc[i]))
+            filenum = filenum + files_per_proc[i]
+            worker = Process(
+                target=self._worker_function,
+                args=(child_connection, proc_list, self._shared_buffer_raw),
+            )
+            worker.start()
+            self._workers.append((worker, parent_connection))
+
+    def keys(self):
+        """Return a list of available variable names"""
+        return self.grid_info["varNames"]
 
     def evolvingVariables(self):
-        """Return a list of names of time-evolving variables
-
-        """
-        return self.evolvingVariableNames
+        """Return a list of names of time-evolving variables"""
+        return self.grid_info["evolvingVariableNames"]
 
     def redistribute(self, npes, nxpe=None, mxg=2, myg=2, include_restarts=True):
         """Create a new set of dump files for npes processors.
@@ -981,14 +1252,8 @@ class BoutOutputs(object):
         old_processor_layout = get_processor_layout(
             DataFile(self._file_list[0]), has_t_dimension=True, mxg=mxg, myg=myg
         )
-        old_nxpe = old_processor_layout.nxpe
-        old_nype = old_processor_layout.nype
-        old_npes = old_processor_layout.npes
-        old_mxsub = old_processor_layout.mxsub
-        old_mysub = old_processor_layout.mysub
         nx = old_processor_layout.nx
         ny = old_processor_layout.ny
-        mz = old_processor_layout.mz
         mxg = old_processor_layout.mxg
         myg = old_processor_layout.myg
 
@@ -1002,7 +1267,8 @@ class BoutOutputs(object):
         mysub = new_processor_layout.mysub
 
         # move existing files to backup directory
-        # don't overwrite backup: os.mkdir will raise exception if directory already exists
+        # don't overwrite backup: os.mkdir will raise exception if directory already
+        # exists
         backupdir = path.join(self._path, "redistribution_backups")
         mkdir(backupdir)
         for f in self._file_list:
@@ -1019,7 +1285,8 @@ class BoutOutputs(object):
                 self._path, this_prefix + str(i) + "." + self._suffix
             )
             if self._suffix.split(".")[-1] in ["nc", "ncdf", "cdl"]:
-                # set format option to DataFile explicitly to avoid creating netCDF3 files, which can only contain up to 2GB of data
+                # set format option to DataFile explicitly to avoid creating netCDF3
+                # files, which can only contain up to 2GB of data
                 outfile_list.append(
                     DataFile(outpath, write=True, create=True, format="NETCDF4")
                 )
@@ -1033,7 +1300,7 @@ class BoutOutputs(object):
             DataFileCache = None
         # read and write the data
         for v in self.varNames:
-            print("processing " + v)
+            print("processing {}".format(v))
             data = collect(
                 v,
                 path=backupdir,
@@ -1072,9 +1339,8 @@ class BoutOutputs(object):
                         # FieldPerp?
                         # check is not perfect, fails if ny=nz
                         raise ValueError(
-                            "Error: Found FieldPerp '"
-                            + v
-                            + "'. This case is not currently handled by BoutOutputs.redistribute()."
+                            "Error: Found FieldPerp '{}'. This case is not currently "
+                            "handled by BoutOutputs.redistribute().".format(v)
                         )
                     outfile.write(
                         v,
@@ -1089,9 +1355,8 @@ class BoutOutputs(object):
                         # evolving Field2D, but this case is not handled
                         # check is not perfect, fails if ny=nx and nx=nt
                         raise ValueError(
-                            "Error: Found evolving Field2D '"
-                            + v
-                            + "'. This case is not currently handled by BoutOutputs.redistribute()."
+                            "Error: Found evolving Field2D '{}'. This case is not "
+                            "currently handled by BoutOutputs.redistribute().".format(v)
                         )
                     outfile.write(
                         v,
@@ -1137,14 +1402,183 @@ class BoutOutputs(object):
                 npes, path=backupdir, nxpe=nxpe, output=self._path, mxg=mxg, myg=myg
             )
 
-    def _collect(self, *args, **kwargs):
-        """Wrapper for collect to pass self._DataFileCache if necessary.
+    def _collect(self, varname):
+        """Wrapper for collect to pass self._DataFileCache if necessary."""
+        if self._parallel:
+            return self._collect_parallel(varname)
 
-        """
         if self._DataFileCaching and self._DataFileCache is None:
             # Need to create the cache
             self._DataFileCache = create_cache(self._path, self._prefix)
-        return collect(*args, datafile_cache=self._DataFileCache, **kwargs)
+        return collect(
+            varname,
+            datafile_cache=self._DataFileCache,
+            path=self._path,
+            prefix=self._prefix,
+            info=self._info,
+            xguards=self._xguards,
+            yguards=self._yguards,
+            tind=self.tind,
+            xind=self.xind,
+            yind=self.yind,
+            zind=self.zind,
+            **self._kwargs,
+        )
+
+    def _collect_parallel(self, varname):
+        tind_auto = self._kwargs.get("tind_auto", False)
+        strict = self._kwargs.get("strict", False)
+        unsupported_kwargs = [
+            k for k in self._kwargs if k not in ("tind_auto", "strict")
+        ]
+        if unsupported_kwargs:
+            raise ValueError(
+                "kwargs {} are not supported when parallel is not False".format(
+                    unsupported_kwargs
+                )
+            )
+
+        if tind_auto:
+            raise ValueError("tind_auto not supported when parallel=True")
+
+        if varname not in self.keys():
+            if strict:
+                raise ValueError("Variable '{}' not found".format(varname))
+            else:
+                varname = findVar(varname, self.keys())
+
+        dimensions = self.grid_info["dimensions"][varname]
+        var_attributes = self.grid_info["attributes"][varname]
+
+        if not ("x" in dimensions or "y" in dimensions or "z" in dimensions):
+            # No spatial dependence - read without using workers to preserve type
+            if "t" in dimensions:
+                return BoutArray(
+                    self._root_file.read(
+                        varname, ranges=[self.tind] + (len(dimensions) - 1) * [None]
+                    ),
+                    attributes=var_attributes,
+                )
+            else:
+                # No time or space dimensions, so no slicing
+                return BoutArray(
+                    self._root_file.read(varname), attributes=var_attributes
+                )
+        elif any(dim not in ("t", "x", "y", "z") for dim in dimensions):
+            raise ValueError(
+                "Dimensions {} of {} contain spatial dimensions but also have "
+                "dimensions that are not 't', 'x', 'y' or 'z'. This is not supported "
+                "by parallel reading. Try reading with parallel=False".format(
+                    dimensions, varname
+                )
+            )
+
+        is_fieldperp = dimensions in (("t", "x", "z"), ("x", "z"))
+
+        # Initialise buffer to zero
+        self._shared_buffer[:] = 0.0
+
+        for worker, connection in self._workers:
+            connection.send((varname, is_fieldperp))
+
+        yindex_global = None
+        fieldperp_yproc = None
+
+        for worker, connection in self._workers:
+            temp_yindex, temp_fieldperp_yproc, temp_var_attributes = connection.recv()
+            if is_fieldperp:
+                (
+                    yindex_global,
+                    fieldperp_yproc,
+                    var_attributes,
+                ) = _check_fieldperp_attributes(
+                    varname,
+                    yindex_global,
+                    temp_yindex,
+                    temp_fieldperp_yproc,
+                    fieldperp_yproc,
+                    var_attributes,
+                    temp_var_attributes,
+                )
+
+        global_slices = []
+        if "t" in dimensions:
+            global_slices.append(slice(None))
+        else:
+            global_slices.append(0)
+        if "x" in dimensions:
+            # Apply any requested step here, after collecting
+            global_slices.append(slice(None, None, self.xind.step))
+        else:
+            global_slices.append(0)
+        if "y" in dimensions:
+            # Apply any requested step here, after collecting
+            global_slices.append(slice(None, None, self.yind.step))
+        else:
+            global_slices.append(0)
+        if "z" in dimensions:
+            global_slices.append(slice(None))
+        else:
+            global_slices.append(0)
+        global_slices = tuple(global_slices)
+
+        return BoutArray(
+            self._shared_buffer[global_slices].copy(), attributes=var_attributes
+        )
+
+    def _worker_function(self, connection, proc_list, shared_buffer_raw):
+        data_files = [DataFile(self._file_list[i]) for i in proc_list]
+        dim_sizes = tuple(self.grid_info["sizes"][d] for d in ("t", "x", "y", "z"))
+        shared_buffer = numpy.reshape(numpy.frombuffer(shared_buffer_raw), dim_sizes)
+        while True:
+            args = connection.recv()
+            if args is None:
+                # Terminate process cleanly
+                for f in data_files:
+                    f.close()
+                connection.close()
+                return 0
+
+            varname, is_fieldperp = args
+
+            yindex_global = None
+            fieldperp_yproc = None
+            var_attributes = None
+
+            for i, f in zip(proc_list, data_files):
+                temp_yindex, temp_var_attributes = _collect_from_one_proc(
+                    i,
+                    f,
+                    varname,
+                    result=shared_buffer,
+                    is_fieldperp=is_fieldperp,
+                    grid_info=self.grid_info,
+                    dimensions=self.grid_info["dimensions"][varname],
+                    tind=self.tind,
+                    xind=self.xind,
+                    yind=self.yind,
+                    zind=self.zind,
+                    xguards=self._xguards,
+                    yguards=self._yguards,
+                    info=self._info,
+                    parallel_read=True,
+                )
+                if is_fieldperp:
+                    (
+                        yindex_global,
+                        fieldperp_yproc,
+                        var_attributes,
+                    ) = _check_fieldperp_attributes(
+                        varname,
+                        yindex_global,
+                        temp_yindex,
+                        i // self.grid_info["nxpe"],
+                        fieldperp_yproc,
+                        var_attributes,
+                        temp_var_attributes,
+                    )
+
+            connection.send((yindex_global, fieldperp_yproc, var_attributes))
 
     def __len__(self):
         return len(self.varNames)
@@ -1159,9 +1593,7 @@ class BoutOutputs(object):
 
         if self._caching:
             if name not in self._datacache.keys():
-                item = self._collect(
-                    name, path=self._path, prefix=self._prefix, **self._kwargs
-                )
+                item = self._collect(name)
                 if self._caching is not True:
                     itemsize = item.nbytes
                     if itemsize > self._datacachemaxsize:
@@ -1177,15 +1609,11 @@ class BoutOutputs(object):
                 return self._datacache[name]
         else:
             # Collect the data from the repository
-            data = self._collect(
-                name, path=self._path, prefix=self._prefix, **self._kwargs
-            )
+            data = self._collect(name)
             return data
 
     def _removeFirstFromCache(self):
-        """Pop the first item from the OrderedDict _datacache
-
-        """
+        """Pop the first item from the OrderedDict _datacache"""
         item = self._datacache.popitem(last=False)
         self._datacachesize -= item[1].nbytes
 
@@ -1198,9 +1626,7 @@ class BoutOutputs(object):
             yield k
 
     def __str__(self, indent=""):
-        """Print a pretty version of the tree
-
-        """
+        """Print a pretty version of the tree"""
         text = ""
         for k in self.varNames:
             text += indent + k + "\n"
